# Copyright (c) 2017-2019 David Steele <dsteele@gmail.com>
#
# SPDX-License-Identifier: GPL-2.0-or-later
# License-Filename: LICENSE

import logging
import os
import sys
import textwrap

import pytest
from mock import Mock

sys.modules["NetworkManager"] = Mock()

from comitup import comitup as ciu  # noqa
from comitup import config  # noqa


@pytest.fixture()
def conf_fxt(tmpdir, monkeypatch):
    path = os.path.join(tmpdir.__str__(), "conffile")

    open(path, "w").write(
        textwrap.dedent(
            """
        ap_name: test
        """
        )
    )

    monkeypatch.setattr("comitup.comitup.config.CONF_PATH", path)

    return path


@pytest.fixture()
def persist_fxt(tmpdir, monkeypatch):
    path = os.path.join(tmpdir.__str__(), "persistfile")

    monkeypatch.setattr("comitup.comitup.config.PERSIST_PATH", path)
    monkeypatch.setattr(
        "comitup.comitup.config.random.randrange", Mock(return_value=1234)
    )

    return path


@pytest.fixture()
def log_fxt(tmpdir, monkeypatch):
    path = os.path.join(tmpdir.__str__(), "logfile")

    monkeypatch.setattr("comitup.comitup.LOG_PATH", path)

    return path


def test_ciu_deflog(log_fxt):
    log = ciu.deflog(logging.INFO)

    log.info("foo")

    txt = open(log_fxt, "r").read()

    assert "INFO" in txt
    assert "foo" in txt
    assert int(txt[:4])


def test_ciu_loadconf(conf_fxt, persist_fxt):
    (conf, data) = config.load_data()
    assert conf.ap_name == "test"
    assert os.path.isfile(persist_fxt)



#!/usr/bin/python3
# Copyright (c) 2017-2019 David Steele <dsteele@gmail.com>
#
# SPDX-License-Identifier: GPL-2.0-or-later
# License-Filename: LICENSE

#
# Copyright 2016-2017 David Steele <steele@debian.org>
# This file is part of comitup
# Available under the terms of the GNU General Public License version 2
# or later
#

import sys

import dbus

from comitup.blink import blink, can_blink  # noqa


class CiuClient(object):
    # Make the dbus functions late binding, to facilitate testing.

    # These are the methods supported, with the corresponding dbus
    # endpoint name.
    methods = {
        "ciu_info": "get_info",
        "ciu_state": "state",
        "ciu_points": "access_points",
        "ciu_connect": "connect",
        "ciu_delete": "delete_connection",
        "ciu_nuke": "nuke",
    }

    def __init__(self):
        self.service = None

    def __getattr__(self, name):
        if name not in self.methods:
            raise AttributeError("Attribute {} not found".format(name))

        try:
            if not self.service:
                bus = dbus.SystemBus()
                self.service = bus.get_object(
                    "com.github.davesteele.comitup",
                    "/com/github/davesteele/comitup",
                )
            func = self.service.get_dbus_method(
                self.methods[name], "com.github.davesteele.comitup"
            )
        except dbus.exceptions.DBusException:
            sys.exit(1)

        # connect as an unbound function
        self.__setattr__(name, func)
        return func



# Copyright (c) 2017-2021 David Steele <dsteele@gmail.com>
#
# SPDX-License-Identifier: GPL-2.0-or-later
# License-Filename: LICENSE
#
# Copyright 2016-2017 David Steele <steele@debian.org>
# This file is part of comitup
# Available under the terms of the GNU General Public License version 2
# or later
#

import configparser
import io
import os
import random
import shutil
from typing import Optional, Tuple

from comitup import persist

PERSIST_PATH: str = "/var/lib/comitup/comitup.json"
CONF_PATH: str = "/etc/comitup.conf"
BOOT_CONF_PATH: str = "/boot/comitup.conf"
SECTION: str = "DEFAULT"

data_cache: Optional[Tuple["Config", persist.persist]] = None


class MyConfigParser(configparser.ConfigParser):
    def getboolean(self, *args, **kwargs):
        mappings = {"y": True, "n": False}

        try:
            val = mappings[self.get(*args, **kwargs).lower()]
        except KeyError:
            val = super().getboolean(*args, **kwargs)

        return val


class Config(object):
    def __init__(self, filename: str, section: str = SECTION, defaults={}):
        self._section: str = section

        self._config = MyConfigParser(defaults=defaults)
        try:
            with open(filename, "r") as fp:
                conf_str = "[%s]\n" % self._section + fp.read()
            conf_fp = io.StringIO(conf_str)
            self._config.read_file(conf_fp)
        except FileNotFoundError:
            pass

    def getboolean(self, tag: str) -> bool:
        return self._config.getboolean(SECTION, tag)

    def __getattr__(self, tag):
        try:
            return self._config.get(self._section, tag)
        except configparser.NoOptionError:
            raise AttributeError


def load_data() -> Tuple[Config, persist.persist]:
    global data_cache

    if not data_cache:
        if os.path.isfile(BOOT_CONF_PATH):
            try:
                dest = shutil.copyfile(BOOT_CONF_PATH, CONF_PATH)
                print("Boot config file copied:", dest)
                os.remove(BOOT_CONF_PATH)
            except Exception:
                print("Error occurred while copying file.")

        conf = Config(
            CONF_PATH,
            defaults={
                "ap_name": "comitup-<nnn>",
                "ap_password": "",
                "web_service": "",
                "service_name": "comitup",
                "external_callback": "/usr/local/bin/comitup-callback",
                "verbose": "0",
                "enable_appliance_mode": "1",
                "primary_wifi_device": "",
                "enable_nuke": "0",
                "ipv6_link_local": 1,
            },
        )

        data = persist.persist(
            PERSIST_PATH,
            {"id": str(random.randrange(1000, 9999))},
        )

        data_cache = (conf, data)

    return data_cache



#!/usr/bin/python3
# Copyright (c) 2017-2019 David Steele <dsteele@gmail.com>
#
# SPDX-License-Identifier: GPL-2.0-or-later
# License-Filename: LICENSE

#
# Copyright 2016-2017 David Steele <steele@debian.org>
# This file is part of comitup
# Available under the terms of the GNU General Public License version 2
# or later
#

import logging
import os
import signal
import socket
import subprocess
from typing import Dict, List, Optional

import dbus
import NetworkManager

from comitup import config, nm

log = logging.getLogger("comitup")

# globals

CLASS_IN = 0x01
TYPE_A = 0x01
TYPE_AAAA = 0x1C
TTL = 5

DBUS_NAME = "org.freedesktop.Avahi"
DBUS_PATH_SERVER = "/"
DBUS_INTERFACE_SERVER = "org.freedesktop.Avahi.Server"
DBUS_INTERFACE_ENTRY_GROUP = "org.freedesktop.Avahi.EntryGroup"
PROTO_UNSPECIFIED = -1
PROTO_INET = 0
PROTO_INET6 = 1
group: Optional[dbus.Interface] = None

# functions


def establish_group() -> None:
    global group

    if not group:
        bus = dbus.SystemBus()

        server = dbus.Interface(
            bus.get_object(DBUS_NAME, DBUS_PATH_SERVER), DBUS_INTERFACE_SERVER
        )

        group = dbus.Interface(
            bus.get_object(DBUS_NAME, server.EntryGroupNew()),
            DBUS_INTERFACE_ENTRY_GROUP,
        )


def encode_dns(name: str) -> bytes:
    components = [x for x in name.split(".") if len(x) > 0]
    fixed_name = ".".join(components)
    return fixed_name.encode("ascii")


def make_a_record(host: str, devindex: int, addr: str) -> None:
    if group:
        group.AddRecord(
            devindex,
            PROTO_INET,
            dbus.UInt32(0),
            encode_dns(host),
            CLASS_IN,
            TYPE_A,
            TTL,
            socket.inet_aton(addr),
        )


def make_aaaa_record(host: str, devindex: int, addr: str) -> None:
    if group:
        group.AddRecord(
            devindex,
            PROTO_INET6,
            dbus.UInt32(0),
            encode_dns(host),
            CLASS_IN,
            TYPE_AAAA,
            TTL,
            socket.inet_pton(socket.AF_INET6, addr),
        )


def string_to_txt_array(strng: str) -> List[bytes]:
    if strng:
        return [dbus.Byte(x) for x in strng.encode()]
    else:
        return []


def string_array_to_txt_array(txt_array: List[str]) -> List[List[bytes]]:
    return [string_to_txt_array(x) for x in txt_array]


def add_service(
    host: str, devindex: int, addr: Optional[str], addr6: Optional[str]
) -> None:
    name = host
    (conf, data) = config.load_data()
    if ".local" in name:
        name = name[: -len(".local")]

    if group:
        group.AddService(
            devindex,
            PROTO_UNSPECIFIED,
            dbus.UInt32(0),
            name,
            "_%s._tcp" % conf.service_name,
            "",
            host,
            dbus.UInt16(9),
            string_array_to_txt_array(
                [
                    "hostname=%s" % host,
                    (
                        "ipaddr=%s" % addr
                        if (addr and addr != "0.0.0.0")
                        else "ipaddr="
                    ),
                    "ip6addr=%s" % addr6 if addr6 else "ip6addr=",
                    "comitup-home=https://davesteele.github.io/comitup/",
                ]
            ),
        )


# public functions


def clear_entries(emphatic=False) -> None:
    global group

    if group and (not group.IsEmpty() or emphatic):
        group.Reset()

        if emphatic:
            group.Free()
            group = None

    establish_group()


def get_interface_mapping() -> Dict[str, int]:
    mapping: Dict[str, int] = {}

    for line in subprocess.check_output(["ip", "addr"]).decode().split("\n"):
        try:
            asc_index, name = line.split(": ")[0:2]
            mapping[name] = int(asc_index)
        except ValueError:
            pass

    return mapping


def add_hosts(hosts: List[str]) -> None:
    establish_group()
    int_mapping = get_interface_mapping()

    devices: Optional[List["NetworkManager.Device"]] = nm.get_devices()

    if devices is None:
        log.error("Null device list returned in add_hosts()")
        return

    if not devices:
        log.error("No devices found in add_hosts()")
        return

    entries: bool = False
    for device in devices:
        name = nm.device_name(device)
        addr = nm.get_active_ip(device)
        addr6 = nm.get_active_ip6(device)
        log.debug("add_hosts: {}, {}".format(name, addr))
        try:
            if name in nm.get_phys_dev_names() and name in int_mapping:
                index = int_mapping[name]

                try:
                    if addr and addr != "0.0.0.0":
                        for host in hosts:
                            log.debug(
                                "Add A record {}-{}-{}".format(
                                    host, index, addr
                                )
                            )
                            make_a_record(host, index, addr)

                        entries = True

                    if addr6:
                        for host in hosts:
                            log.debug(
                                "Add AAAA record {}-{}-{}".format(
                                    host, index, addr6
                                )
                            )
                            make_aaaa_record(host, index, addr6)

                        entries = True

                    if addr6 or (addr and addr != "0.0.0.0"):
                        log.debug(
                            "Add service {}, {}, {}-{}".format(
                                host, index, addr, addr6
                            )
                        )
                        add_service(hosts[0], index, addr, addr6)

                except Exception:
                    log.error("Exception encountered adding avahi record")
                    clear_entries(emphatic=True)
                    entries = False
        except NetworkManager.ObjectVanished as e:
            log.error(
                "Unrecoverable NetworkManager Error - exiting - {}".format(
                    str(e)
                )
            )
            os.kill(os.getpid(), signal.SIGTERM)

    if group and entries:
        group.Commit()


def check_mdns(hosts: List[str]) -> None:
    reset = False

    log.debug("Checking mdns")

    try:
        socket.gethostbyname(hosts[0])
    except socket.gaierror:
        reset = True

    if reset:
        log.error("Mdns check failed - resetting records and services.")
        clear_entries(emphatic=True)
        add_hosts(hosts)


if __name__ == "__main__":
    add_hosts(["comitup-1111.local", "comitup.local"])

    while True:
        pass



#!/usr/bin/python3
#
# Copyright (c) 2022 David Steele <dsteele@gmail.com>
#
# SPDX-License-Identifier: GPL-2.0-or-later
# License-Filename: LICENSE
#

"""
devtest.py

This creates a  virtual environment, and runs a number of test environments
against the comitup code.

The venv is persistent, and the tests run in parallel, so this is much quicker
than tox or nox.
"""

import subprocess
import sys
import venv
from concurrent.futures import ThreadPoolExecutor
from pathlib import Path
from typing import List

envpath: Path = Path(__file__).resolve().parent / ".devenv"
pythonpath: str = str(envpath / "bin" / "python")


pkgs: List[str] = [
    "pytest",
    "mypy",
    "flake8",
    "black",
    "isort",
    "cachetools",
    "flask",
    "types-mock",
    "types-tabulate",
    "types-pkg_resources",
    "types-Flask",
    "types-cachetools",
]

targets: str = "comitup web cli test devtest.py setup.py"


def mkcmd(cmd: str) -> List[str]:
    return [str(pythonpath), "-m"] + cmd.split()


def run(cmd: str) -> subprocess.CompletedProcess:
    cp = subprocess.run(
        mkcmd(cmd), stdout=subprocess.PIPE, stderr=subprocess.STDOUT
    )

    return cp


print("# Tests starting")

if not envpath.exists():
    print("# Creating virtual environment")

    virtenv = venv.EnvBuilder(
        system_site_packages=True, symlinks=True, with_pip=True
    )
    virtenv.create(str(envpath))

    print("# Installing packages")

    run("pip install " + " ".join(pkgs))


tests: List[str] = [
    "black --check {}".format(targets),
    "isort --check {}".format(targets),
    "mypy {}".format(targets),
    "flake8 {}".format(targets),
    "pytest",
]

executor = ThreadPoolExecutor(max_workers=5)

fail = False
for result in executor.map(lambda x: run(x), tests):
    print("#####################################")
    print("# Running {}".format(" ".join(result.args)))
    print(result.stdout.decode())
    print("#####################################")
    print()
    if result.returncode:
        fail = True

if fail:
    print("# ERROR(S) ENCOUNTERED")
    sys.exit(1)

print("# Tests complete")



# Copyright (c) 2022 David Steele <dsteele@gmail.com>
#
# SPDX-License-Identifier: GPL-2.0-or-later
# License-Filename: LICENSE
#

"""
blink.py

Add support to blink the green led on a Raspberry pi one time, for a half sec.

https://mlagerberg.gitbooks.io/raspberry-pi/content/5.2-leds.html
"""

import re
import time
from pathlib import Path

brightPath: Path = Path("/sys/class/leds/led0/brightness")
triggerPath: Path = Path("/sys/class/leds/led0/trigger")
modelPath: Path = Path("/sys/firmware/devicetree/base/model")


def onval() -> str:
    """A "1" turns on the led."""
    return "1"


def offval() -> str:
    """Value to turn the led off."""
    return "0" if onval() == "1" else "1"


def can_blink() -> bool:
    """Is this a Pi with a blinkable green led?"""
    if brightPath.exists() and triggerPath.exists():
        return True

    return False


def get_trigger() -> str:
    """Save the current led trigger, for later restoration."""
    text: str = triggerPath.read_text()

    match = re.search(r"\[(.+)\]", text)

    mode: str
    if match:
        mode = match.group(1)
    else:
        mode = "none"

    return mode


def set_trigger(trigger: str) -> None:
    """Set the green led trigger."""
    triggerPath.write_text(trigger)


def blink(times: int = 1) -> None:
    """Blink the green led once."""
    if can_blink():
        oldtrig = get_trigger()

        set_trigger("gpio")
        brightPath.write_text(offval())

        for _ in range(times):
            time.sleep(0.25)
            brightPath.write_text(onval())
            time.sleep(0.5)
            brightPath.write_text(offval())
            time.sleep(0.25)

        set_trigger(oldtrig)


if __name__ == "__main__":
    print(can_blink())
    print(get_trigger())
    blink(3)



import pytest
from mock import Mock, patch

from comitup import mdns

# Copyright (c) 2017-2019 David Steele <dsteele@gmail.com>
#
# SPDX-License-Identifier: GPL-2.0-or-later
# License-Filename: LICENSE


@pytest.fixture()
def avahi_fxt(monkeypatch, request):
    monkeypatch.setattr("comitup.mdns.dbus.Interface", Mock())
    monkeypatch.setattr("comitup.mdns.dbus.SystemBus", Mock())
    monkeypatch.setattr("comitup.mdns.log", Mock())
    monkeypatch.setattr("comitup.mdns.config.persist", Mock())

    save_group = mdns.group
    mdns.group = Mock()

    yield None

    mdns.group = save_group


def test_avahi_null(avahi_fxt):
    pass


def test_avahi_establish_group(avahi_fxt):
    old_group = mdns.group
    mdns.group = None
    mdns.establish_group()
    assert mdns.group is not None
    mdns.group = old_group


def test_avahi_make_a_record(avahi_fxt):
    mdns.make_a_record("host", 1, "1.2.3.4")
    assert mdns.group.AddRecord.called  # type: ignore


def test_avahi_add_service(avahi_fxt):
    mdns.add_service("host", 1, "1.2.3.4", "::1")
    assert mdns.group.AddService.called  # type: ignore


@patch("comitup.mdns.establish_group", Mock())
def test_avahi_clear_entries(avahi_fxt):
    isempty = Mock(return_value=False)
    mdns.group = Mock()
    mdns.group.IsEmpty = isempty

    oldgroup = mdns.group

    mdns.clear_entries()

    assert isempty.called
    assert oldgroup.Reset.called
    assert not mdns.log.called  # type: ignore


@pytest.mark.parametrize(
    "dns_in, dns_out",
    (
        ("a.b.c", "a.b.c".encode()),
        ("A.B.C", "A.B.C".encode()),
        ("a..b", "a.b".encode()),
        ("a.b.", "a.b".encode()),
    ),
)
def test_avahi_encode_dns(dns_in, dns_out):
    assert dns_out == mdns.encode_dns(dns_in)


@patch("comitup.mdns.log.warn")
def test_avahi_clear_fail(warn, avahi_fxt):
    mdns.group = None
    mdns.clear_entries()



# Copyright (c) 2017-2019 David Steele <dsteele@gmail.com>
#
# SPDX-License-Identifier: GPL-2.0-or-later
# License-Filename: LICENSE
#
# Copyright 2016-2017 David Steele <steele@debian.org>
# This file is part of comitup
# Available under the terms of the GNU General Public License version 2
# or later
#

import logging

from .sysd import sd_start_unit, sd_stop_unit

log: logging.Logger = logging.getLogger("comitup")

COMITUP_SERVICE: str = "comitup-web.service"

web_service: str = ""


def start_service(service: str) -> None:
    log.debug("starting %s web service", service)
    sd_start_unit(service, "replace")


def stop_service(service: str) -> None:
    log.debug("stopping %s web service", service)
    sd_stop_unit(service, "replace")


callmatrix = {
    ("HOTSPOT", "start"): (lambda: stop_service, lambda: web_service),
    ("HOTSPOT", "pass"): (lambda: start_service, lambda: COMITUP_SERVICE),
    ("CONNECTING", "start"): (lambda: stop_service, lambda: COMITUP_SERVICE),
    ("CONNECTED", "start"): (lambda: start_service, lambda: web_service),
}


def state_callback(state: str, action: str) -> None:
    try:
        (fn_fact, svc_fact) = callmatrix[(state, action)]
    except KeyError:
        return

    if svc_fact():
        fn_fact()(svc_fact())


def init_webmgr(web_svc: str) -> None:
    global web_service

    stop_service(COMITUP_SERVICE)

    web_service = web_svc

    log.debug("web service is %s" % web_service)



#!/usr/bin/python3
# Copyright (c) 2019 David Steele <dsteele@gmail.com>
#
# SPDX-License-Identifier: GPL-2.0-or-later
# License-Filename: LICENSE

import logging
import subprocess
import time

log = logging.getLogger("comitup")

# The AP hotspot is eventually failing. Restarting wpa_supplicant recovers it.
# I haven't found a way to detect that the wpa_supplicant is not working
# properly without involving the outside world.
#
# The minimal workaround fix is to call disconnect, then reconnect from
# wpa_cli.

last_kick_time: float = 0
kick_period_secs: int = 5 * 60


def needs_kick(devstring: str) -> bool:
    return (time.time() - kick_period_secs) > last_kick_time


def kick_wpa(devstring: str) -> None:
    global last_kick_time

    log.debug("Kicking wpa on {}".format(devstring))

    for wpa_cmd in ("disconnect", "reconnect"):
        shell_cmd = "/sbin/wpa_cli -i {0} {1}".format(devstring, wpa_cmd)
        subprocess.run(shell_cmd.split())

    last_kick_time = time.time()


def check_wpa(devstring: str) -> None:
    if needs_kick(devstring):
        kick_wpa(devstring)



# Copyright (c) 2019 David Steele <dsteele@gmail.com>
#
# SPDX-License-Identifier: GPL-2.0-or-later
# License-Filename: LICENSE
#

import logging
import os
import signal
import subprocess
import time
from pathlib import Path

from comitup import modemgr

log = logging.getLogger("comitup")

hotspot_config: str = "/usr/share/comitup/dns/dns-hotspot.conf"
connected_config: str = "/usr/share/comitup/dns/dns-connected.conf"

pidpath: Path = Path("/var/run/comitup-dns")

callmatrix = {
    ("HOTSPOT", "pass", "router"): (lambda: run_dns, lambda: hotspot_config),
    ("CONNECTING", "start", "router"): (
        lambda: run_dns,
        lambda: connected_config,
    ),
    ("HOTSPOT", "pass", "single"): (lambda: run_dns, lambda: hotspot_config),
    ("CONNECTING", "start", "single"): (lambda: run_dns, lambda: ""),
}


def kill_dns(ppath: Path, sig: int) -> None:
    try:
        pid: int = int(pidpath.read_text().strip())
        os.kill(pid, sig)
        os.waitpid(pid, 0)
    except (
        ValueError,
        ProcessLookupError,
        FileNotFoundError,
        ChildProcessError,
    ):
        pass


def run_dns(confpath: str) -> None:
    log.debug("Running dnsmasq using {}".format(confpath))

    kill_dns(pidpath, signal.SIGTERM)

    if os.path.exists(confpath):
        dev = modemgr.get_ap_device().Interface

        cmd = "dnsmasq --conf-file={0} --interface={1}".format(confpath, dev)

        for _ in range(5):
            cp = subprocess.run(cmd.split())
            if cp.returncode == 0:
                break
            time.sleep(0.1)


def state_callback(state: str, action: str) -> None:
    try:
        (fn_fact, svc_fact) = callmatrix[(state, action, modemgr.get_mode())]
    except KeyError:
        return

    fn_fact()(svc_fact())



# Copyright (c) 2017-2019 David Steele <dsteele@gmail.com>
#
# SPDX-License-Identifier: GPL-2.0-or-later
# License-Filename: LICENSE
#
# Copyright 2017 David Steele <steele@debian.org>
# This file is part of comitup
# Available under the terms of the GNU General Public License version 2
# or later
#

import logging
import subprocess
from typing import List, Optional

from comitup import modemgr, nm, routemgr

# fmt: off
start_cmds: List[str] = [
    # HOTSPOT rules
    "iptables -w -N COMITUP-OUT",
    "iptables -w -A COMITUP-OUT "
      "-p icmp --icmp-type destination-unreachable -j DROP",  # noqa
    "iptables -w -A COMITUP-OUT "
      "-p icmp --icmp-type port-unreachable -j DROP",
    "iptables -w -A COMITUP-OUT -j RETURN",
    "iptables -w -I OUTPUT -o {ap} -j COMITUP-OUT",
]

end_cmds: List[str] = [
    # Clear HOTSPOT rules
    "iptables -w -D OUTPUT -o {ap} -j COMITUP-OUT >/dev/null 2>&1",
    "iptables -w -F COMITUP-OUT >/dev/null 2>&1",
    "iptables -w -X COMITUP-OUT >/dev/null 2>&1",
]

appliance_cmds: List[str] = [
    "iptables -w -t nat -N COMITUP-FWD",
    "iptables -w -t nat -A COMITUP-FWD -o {link} -j MASQUERADE",
    "iptables -w -t nat -A COMITUP-FWD -j RETURN",
    "iptables -w -t nat -A POSTROUTING -j COMITUP-FWD",
    "echo 1 > /proc/sys/net/ipv4/ip_forward",
]

appliance_clear: List[str] = [
    "iptables -w -t nat -D POSTROUTING -j COMITUP-FWD >/dev/null 2>&1",
    "iptables -w -t nat -F COMITUP-FWD >/dev/null 2>&1",
    "iptables -w -t nat -X COMITUP-FWD >/dev/null 2>&1",
]
# fmt: on

log: logging.Logger = logging.getLogger("comitup")


def run_cmds(cmds: List[str]) -> None:
    linkdev = nm.device_name(modemgr.get_link_device())
    apdev = nm.device_name(modemgr.get_ap_device())
    for cmd in cmds:
        subprocess.call(cmd.format(link=linkdev, ap=apdev), shell=True)


def state_callback(state: str, action: str) -> None:
    if (state, action) == ("HOTSPOT", "start"):
        log.debug("Running iptables commands for HOTSPOT")

        run_cmds(end_cmds)
        run_cmds(start_cmds)

        if modemgr.get_mode() == modemgr.MULTI_MODE:
            run_cmds(appliance_clear)

        log.debug("Done with iptables commands for HOTSPOT")

    elif (state, action) == ("CONNECTED", "start"):
        log.debug("Running iptables commands for CONNECTED")
        run_cmds(end_cmds)

        if modemgr.get_mode() == modemgr.MULTI_MODE:
            run_cmds(appliance_clear)
            run_cmds(appliance_cmds)
            defaultdev: Optional[str] = routemgr.defroute_dev()
            apdev: str = modemgr.get_ap_device().Interface
            if defaultdev and defaultdev != apdev:
                run_cmds(
                    [
                        "iptables -w -t nat -I COMITUP-FWD -o "
                        "{} -j MASQUERADE".format(defaultdev)
                    ],
                )

        log.debug("Done with iptables commands for CONNECTED")


def init_iptmgr() -> None:
    pass


def main():
    print("applying rules")
    run_cmds(start_cmds)

    input("Press Enter to continue...")

    run_cmds(end_cmds)


if __name__ == "__main__":
    main()



# Copyright (c) 2017-2019 David Steele <dsteele@gmail.com>
#
# SPDX-License-Identifier: GPL-2.0-or-later
# License-Filename: LICENSE

import importlib
from collections import namedtuple

import dbus.service
import pytest
from mock import Mock


def nullwrapper(*args, **kwargs):
    def _nullwrapper(fn):
        def wrapper(*wargs, **wkwargs):
            return fn(*wargs, **wkwargs)

        return wrapper

    return _nullwrapper


dbus.service.method = nullwrapper

sm = importlib.import_module("comitup.statemgr")


@pytest.fixture()
def statemgr_fxt(monkeypatch, request):
    monkeypatch.setattr("dbus.service.BusName", Mock())
    monkeypatch.setattr("dbus.service.Object", Mock())
    monkeypatch.setattr("dbus.service.Object", Mock())

    save_state = sm.states.com_state
    save_conn = sm.states.connection

    def fin():
        sm.states.com_state = save_state
        sm.states.connection = save_conn

    request.addfinalizer(fin)

    sm.states.com_state = "CONNECTED"
    sm.states.connection = "connection"


def test_sm_none(statemgr_fxt):
    pass


def test_sm_state(statemgr_fxt):
    obj = sm.Comitup()
    assert obj.state() == ["CONNECTED", "connection"]


@pytest.fixture()
def ap_name_fxt(monkeypatch):
    monkeypatch.setattr("socket.gethostname", Mock(return_value="host"))

    return None


Case = namedtuple("Case", ["input", "out"])


@pytest.mark.parametrize(
    "case",
    [
        Case("comitup-<nnn>", "comitup-123"),
        Case("<n>", "1"),
        Case("<nn>", "12"),
        Case("<nnn>", "123"),
        Case("<nnnn>", "1234"),
        Case("<nnnnn>", "<nnnnn>"),
        Case("A<n>", "A1"),
        Case("<n>A", "1A"),
        Case("<hostname>", "host"),
        Case("<hostname>-<n>", "host-1"),
        Case("<bogus>", "<bogus>"),
    ],
)
def test_expand_ap(ap_name_fxt, case):
    assert sm.expand_ap(case.input, "1234") == case.out



#!/usr/bin/python3

# Copyright (c) 2021 David Steele <dsteele@gmail.com>
#
# SPDX-License-Identifier: GPL-2.0-or-later
# License-Filename: LICENSE

import re
import subprocess
from typing import Optional


def defroute_dev() -> Optional[str]:
    """Return the name of the interface holding the default route."""

    cp = subprocess.run(
        "ip route".split(), stdout=subprocess.PIPE, encoding="utf-8"
    )
    fstline = cp.stdout.split("\n")[0]
    match = re.search("^default.+dev ([^ ]+)", fstline)

    if match:
        return match.group(1)

    return None


if __name__ == "__main__":
    print("Default route device is", defroute_dev())



#!/usr/bin/python3
# Copyright (c) 2017-2019 David Steele <dsteele@gmail.com>
#
# SPDX-License-Identifier: GPL-2.0-or-later
# License-Filename: LICENSE

#
# Copyright 2016-2017 David Steele <steele@debian.org>
# This file is part of comitup
# Available under the terms of the GNU General Public License version 2
# or later
#

import logging
from functools import partial
from typing import Callable, List, Optional

import dbus
import NetworkManager
from dbus.mainloop.glib import DBusGMainLoop
from gi.repository.GLib import MainLoop, timeout_add

if __name__ == "__main__":
    DBusGMainLoop(set_as_default=True)

if __name__ == "__main__":
    import os
    import sys

    fullpath = os.path.abspath(__file__)
    parentdir = "/".join(fullpath.split("/")[:-2])
    sys.path.insert(0, parentdir)

from comitup import modemgr  # noqa
from comitup import nm  # noqa

log: logging.Logger = logging.getLogger("comitup")

bus: dbus.SystemBus = dbus.SystemBus()

monitored_dev: Optional[NetworkManager.Device] = None
ap_device: Optional[NetworkManager.Device] = None
second_device_name: Optional[str] = None

nm_dev_connect: Optional[Callable[[], None]] = None
nm_dev_fail: Optional[Callable[[], None]] = None

PASS_STATES: List[int] = [
    # NetworkManager.NM_DEVICE_STATE_IP_CHECK,
    NetworkManager.NM_DEVICE_STATE_ACTIVATED
]
BASE_FAIL_STATES: List[int] = [NetworkManager.NM_DEVICE_STATE_FAILED]
ENHANCED_FAIL_STATES: List[int] = [NetworkManager.NM_DEVICE_STATE_DISCONNECTED]
FAIL_STATES: List[int] = BASE_FAIL_STATES


def base_fail_states() -> None:
    global FAIL_STATES

    FAIL_STATES = BASE_FAIL_STATES


def enhance_fail_states() -> None:
    global FAIL_STATES

    FAIL_STATES = BASE_FAIL_STATES + ENHANCED_FAIL_STATES


def disable() -> None:
    global monitored_dev, nm_dev_connect, nm_dev_fail

    monitored_dev = None

    nm_dev_connect = None
    nm_dev_fail = None


def enable(
    dev: NetworkManager.Device,
    connect_fn: Callable[[], None],
    fail_fn: Callable[[], None],
    state_id: int,
) -> None:
    global monitored_dev, nm_dev_connect, nm_dev_fail

    base_fail_states()

    monitored_dev = None

    nm_dev_connect = partial(connect_fn, state_id)
    nm_dev_fail = partial(fail_fn, state_id)

    monitored_dev = dev


def send_cb(cb: Callable[[], None], reason) -> None:
    def cb_to(cb, reason):
        cb(reason)
        return False

    timeout_add(0, cb_to, cb, reason)


def ap_changed_state(state, oldstate, reason, *args) -> None:
    log.debug(
        "nmm - primary state {}, was {}, reason {}".format(
            state, oldstate, reason
        )
    )
    if state in PASS_STATES:
        log.debug("nmm - primary pass")
        if nm_dev_connect:
            send_cb(nm_dev_connect, reason)
    elif state in FAIL_STATES:
        log.debug("nmm - primary fail")
        if nm_dev_fail:
            send_cb(nm_dev_fail, reason)


def second_changed_state(state, oldstate, reason, *args) -> None:
    log.debug(
        "nmm - secondary state {}, was {}, reason {}".format(
            state, oldstate, reason
        )
    )
    if state in PASS_STATES:
        log.debug("nmm - secondary pass")
        if nm_dev_connect:
            send_cb(nm_dev_connect, reason)
    elif state in FAIL_STATES:
        log.debug("nmm - secondary fail")
        if nm_dev_fail:
            send_cb(nm_dev_fail, reason)


def any_changed_state(state: int, *args) -> None:
    from comitup import mdns
    from comitup.states import dns_names

    interesting_states = PASS_STATES + FAIL_STATES

    def reset_mdns():
        mdns.clear_entries()
        mdns.add_hosts(dns_names)

    if state in interesting_states:
        timeout_add(0, reset_mdns)


def set_device_listeners(
    ap_dev: NetworkManager.Device,
    second_dev: NetworkManager.Device,
) -> None:
    global ap_device, second_device_name

    if ap_device is None:
        log.debug("nmm - Setting primary listener for {}".format(ap_dev))
        ap_device = ap_dev
        device_listener = bus.add_signal_receiver(
            ap_changed_state,
            signal_name="StateChanged",
            dbus_interface="org.freedesktop.NetworkManager.Device",
            path=nm.get_device_path(ap_dev),
        )
        log.debug("Listener is {}".format(device_listener))

    if (
        second_device_name != second_dev.Interface
        and ap_dev.Interface != second_dev.Interface
    ):
        log.debug("nmm - Setting 2nd listener for {}".format(second_dev))
        second_device_name = second_dev.Interface
        device_listener = bus.add_signal_receiver(
            second_changed_state,
            signal_name="StateChanged",
            dbus_interface="org.freedesktop.NetworkManager.Device",
            path=nm.get_device_path(second_dev),
        )
        log.debug("Listener is {}".format(device_listener))

    device_listener = bus.add_signal_receiver(
        any_changed_state,
        signal_name="StateChanged",
        dbus_interface="org.freedesktop.NetworkManager.Device",
        path=None,
    )


def init_nmmon() -> None:
    set_device_listeners(modemgr.get_ap_device(), modemgr.get_link_device())


def main():
    handler = logging.StreamHandler(stream=None)
    log.addHandler(handler)
    log.setLevel(logging.DEBUG)

    log.info("starting")

    init_nmmon()

    def up():
        print("wifi up")

    def down():
        print("wifi down")

    enable(modemgr.get_ap_device(), up, down)

    loop = MainLoop()
    loop.run()


if __name__ == "__main__":
    main()



# Copyright (c) 2017-2019 David Steele <dsteele@gmail.com>
#
# SPDX-License-Identifier: GPL-2.0-or-later
# License-Filename: LICENSE
#
# Copyright 2016-2017 David Steele <steele@debian.org>
# This file is part of comitup
# Available under the terms of the GNU General Public License version 2
# or later
#

import sys

from setuptools import setup

sys.path.append(".")
from comitup import __version__  # noqa

setup(
    name="comitup",
    packages=["comitup", "web", "cli"],
    version=__version__.__version__,
    description="Remotely manage wifi connections on a headless computer",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Web Environment",
        "Framework :: Flask",
        "Intended Audience :: End Users/Desktop",
        "License :: OSI Approved "
        + ":: GNU General Public License v2 or later (GPLv2+)",
        "Natural Language :: English",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python",
        "Topic :: System :: Networking",
    ],
    entry_points={
        "console_scripts": [
            "comitup-cmd=comitup.comitup:main",
            "comitup-cli=cli.comitupcli:main",
            "comitup-web=web.comitupweb:main",
        ],
    },
    options={
        "build_scripts": {
            "executable": "/usr/bin/python3",
        },
    },
    data_files=[
        ("/etc", ["conf/comitup.conf"]),
        ("/usr/share/dbus-1/system.d", ["conf/comitup-dbus.conf"]),
        (
            "/web/templates",
            [
                "web/templates/index.html",
                "web/templates/connect.html",
                "web/templates/confirm.html",
            ],
        ),
        (
            "/web/templates/css",
            [
                "web/templates/css/uikit.css",
                "web/templates/css/uikit-rtl.css",
                "web/templates/css/uikit.min.css",
                "web/templates/css/uikit-rtl.min.css",
            ],
        ),
        (
            "/web/templates/js",
            [
                "web/templates/js/uikit",
                "web/templates/js/uikit-icons",
                "web/templates/js/blink.js",
            ],
        ),
        (
            "/web/templates/images",
            [
                "web/templates/images/ledon.gif",
                "web/templates/images/ledoff.gif",
            ],
        ),
        (
            "/dns",
            [
                "conf/dns-hotspot.conf",
                "conf/dns-connected.conf",
            ],
        ),
    ],
    install_requires=[
        "cachetools",
        "jinja2",
        "pygobject",
        "flask",
        "python-networkmanager",
        "pycairo",
    ],
    setup_requires=["pytest-runner"],
    tests_require=["pytest", "mock"],
    author="David Steele",
    author_email="steele@debian.org",
    url="https://davesteele.github.io/comitup/",
)



# Copyright (c) 2022 David Steele <dsteele@gmail.com>
#
# SPDX-License-Identifier: GPL-2.0-or-later
# License-Filename: LICENSE
#

import subprocess

import nox

pkgs = [
    "libcairo2-dev",
    "gobject-introspection",
    "libgirepository1.0-dev",
    "python3-dev",
    "libdbus-glib-1-dev",
    "libdbus-1-dev",
]

deps = [
    "pytest",
    "mock",
    "dbus-python",
    "python-networkmanager",
    "flask",
    "pygobject",
    "cachetools",
]


def missing_pkg(pkg):
    cmd = "dpkg -l {} > /dev/null".format(pkg)
    return subprocess.run(cmd, shell=True).returncode != 0


@nox.session()
def test(session):
    missings = [x for x in pkgs if missing_pkg(x)]
    if missings:
        session.error("Missing packages: %s" % format(" ".join(missings)))

    for pkg in deps:
        session.install(pkg)

    session.run("python", "-m", "pytest")


@nox.session()
def flake8(session):
    session.install("flake8")
    session.run("python", "-m", "flake8", "setup.py", "cli", "comitup", "web", "test")


@nox.session()
def mypy(session):
    session.install(
        "mypy",
        "types-mock",
        "types-tabulate",
        "types-pkg_resources",
        "types-Flask",
        "types-cachetools",
    )

    session.run("python", "-m", "mypy", "cli", "comitup", "web", "test")



#!/usr/bin/python3
# Copyright (c) 2017-2019 David Steele <dsteele@gmail.com>
#
# SPDX-License-Identifier: GPL-2.0-or-later
# License-Filename: LICENSE

#
# Copyright 2016-2017 David Steele <steele@debian.org>
# This file is part of comitup
# Available under the terms of the GNU General Public License version 2
# or later
#

import logging
import os
import re
import socket
import subprocess
import sys
from typing import TYPE_CHECKING, Callable, Dict, List, Optional

import dbus
import dbus.service

from comitup import iwscan, nuke

if TYPE_CHECKING:
    from comitup.config import Config
    from comitup.persist import persist

sys.path.append("/usr/share/comitup")

import time  # noqa

from dbus.mainloop.glib import DBusGMainLoop  # noqa
from gi.repository.GLib import MainLoop, timeout_add  # noqa

DBusGMainLoop(set_as_default=True)

from comitup import modemgr  # noqa
from comitup import nm  # noqa
from comitup import states  # noqa
from comitup.__version__ import __version__  # noqa

comitup_path: str = "/com/github/davesteele/comitup"

comitup_int: str = "com.github.davesteele.comitup"

log: logging.Logger = logging.getLogger("comitup")


com_obj: Optional["Comitup"] = None
conf: Optional["Config"] = None
data: Optional["persist"] = None

apcache: List[Dict[str, str]] = []
cachetime: float = 0


class Comitup(dbus.service.Object):
    def __init__(self):
        bus_name = dbus.service.BusName(comitup_int, bus=dbus.SystemBus())
        dbus.service.Object.__init__(self, bus_name, comitup_path)

    @dbus.service.method(comitup_int, in_signature="", out_signature="aa{ss}")
    def access_points(self) -> List[Dict[str, str]]:
        global apcache, cachetime

        if time.time() - cachetime > 5:
            cachetime = time.time()  # keep anyone else from processing
            aps = iwscan.candidates()
            aps = [x for x in aps if x["ssid"] != states.hotspot_name]
            apcache = aps

            # set a timeout, if we got something
            if len(apcache):
                cachetime = time.time()  # cache time actually starts now
            else:
                cachetime = 0

        return apcache

    @dbus.service.method(comitup_int, in_signature="", out_signature="ss")
    def state(self):
        return [states.com_state, states.connection]

    @dbus.service.method(comitup_int, in_signature="ss", out_signature="")
    def connect(self, ssid: str, password: str) -> None:
        def to_fn(ssid, password):
            if nm.get_connection_by_ssid(ssid):
                nm.del_connection_by_ssid(ssid)

            nm.make_connection_for(
                ssid, password, link_local=conf.getboolean("ipv6_link_local")
            )

            states.set_state("CONNECTING", [ssid, ssid])
            return False

        timeout_add(1, to_fn, ssid, password)

    @dbus.service.method(comitup_int, in_signature="", out_signature="")
    def delete_connection(self) -> None:
        def to_fn():
            ssid = nm.get_active_ssid(modemgr.get_link_device())
            nm.del_connection_by_ssid(ssid)
            states.set_state("HOTSPOT")
            return False

        timeout_add(1, to_fn)

    @dbus.service.method(comitup_int, in_signature="", out_signature="a{ss}")
    def get_info(self) -> Dict[str, str]:
        if not conf or not data:
            sys.exit(1)
        return get_info(conf, data)

    @dbus.service.method(comitup_int, in_signature="", out_signature="")
    def nuke(self) -> None:
        def do_nuke():
            nuke.nuke()

        timeout_add(10, do_nuke)


def get_info(conf: "Config", data: "persist") -> Dict[str, str]:
    if not conf or not data:
        sys.exit(1)

    info = {
        "version": __version__,
        "apname": expand_ap(conf.ap_name, data.id),
        "hostnames": ";".join(get_hosts(conf, data)),
        "imode": modemgr.get_mode(),
    }

    return info


def expand_ap(ap_name: str, id: str) -> str:
    returnval = ap_name

    for num in range(5):
        returnval = re.sub("<{}>".format("n" * num), id[:num], returnval)

    returnval = re.sub("<hostname>", socket.gethostname(), returnval)

    return returnval


def get_hosts(conf: "Config", data: "persist") -> List[str]:
    return [
        "{}.local".format(expand_ap(conf.ap_name, data.id)),
    ]


def external_callback(state: str, action: str) -> None:
    if action != "start":
        return

    script: str = conf.external_callback  # type: ignore

    if not os.path.isfile(script):
        return

    if not os.access(script, os.X_OK):
        log.error("Callback script %s is not executable" % script)
        return

    log.debug("Calling External callback")

    def demote(uid, gid):
        def dodemote():
            os.setgroups([])
            os.setgid(gid)
            os.setuid(uid)

        return dodemote

    stats = os.stat(script)

    with open(os.devnull, "w") as nul:
        subprocess.call(
            [script, state],
            stdout=nul,
            stderr=subprocess.STDOUT,
            preexec_fn=demote(stats.st_uid, stats.st_gid),
        )


def init_state_mgr(
    gconf: "Config",
    gdata: "persist",
    callbacks: List[Callable[[str, str], None]],
) -> None:
    global com_obj, conf, data

    conf, data = (gconf, gdata)

    states.init_states(
        get_hosts(conf, data),
        callbacks + [external_callback],
        conf.ap_password,
    )
    com_obj = Comitup()


def main():
    handler = logging.StreamHandler(stream=None)
    log.addHandler(handler)
    log.setLevel(logging.DEBUG)

    log.info("starting")

    init_state_mgr("comitup.local", "comitup-1111.local")
    states.set_state("HOTSPOT", timeout=5)

    loop = MainLoop()
    loop.run()


if __name__ == "__main__":
    main()



# Copyright (c) 2017-2019 David Steele <dsteele@gmail.com>
#
# SPDX-License-Identifier: GPL-2.0-or-later
# License-Filename: LICENSE

import os
import shutil
import tempfile

import pytest

from comitup.persist import persist


@pytest.fixture(scope="module")
def dir_fxt(request):
    dir = tempfile.mkdtemp()

    def fin():
        shutil.rmtree(dir)

    request.addfinalizer(fin)

    return dir


count = 0


@pytest.fixture()
def jsonpath(request, dir_fxt):
    global count

    path = os.path.join(dir_fxt, "persist%d.json" % count)
    count += 1

    return path


def test_persist_is_dict(jsonpath):
    mydict = persist(jsonpath)
    mydict["a"] = "b"
    assert mydict["a"] == "b"


def test_persist_default(jsonpath):
    mydict = persist(jsonpath, {"a": "b"})
    assert mydict["a"] == "b"


def test_persist_default_persists(jsonpath):
    persist(jsonpath, {"a": "b"})
    new = persist(jsonpath)
    assert new["a"] == "b"


def test_persist_override_default(jsonpath):
    persist(jsonpath, {"a": "b"})
    new = persist(jsonpath, {"a": "c"})
    assert new["a"] == "b"


def test_persist_override_default2(jsonpath):
    mydict = persist(jsonpath, {"a": "a"})
    mydict["a"] = "b"
    new = persist(jsonpath, {"a": "c"})
    assert new["a"] == "b"


def test_persist_update(jsonpath):
    mydict = persist(jsonpath, {"a": "a"})
    mydict.update({"a": "b"})
    new = persist(jsonpath, {"a": "c"})
    assert new["a"] == "b"


def test_persist_setdefault(jsonpath):
    mydict = persist(jsonpath)
    mydict.setdefault("a", "b")
    new = persist(jsonpath, {"a": "c"})
    assert new["a"] == "b"


def test_persist_setattr(jsonpath):
    mydict = persist(jsonpath)
    mydict.a = "b"
    assert mydict["a"] == "b"


def test_persist_getattr(jsonpath):
    mydict = persist(jsonpath)
    mydict["a"] = "b"
    assert mydict.a == "b"


def test_persist_persist_setattr(jsonpath):
    mydict = persist(jsonpath)
    mydict.a = "b"
    new = persist(jsonpath)
    assert new["a"] == "b"


def test_persist_persist_getattr(jsonpath):
    mydict = persist(jsonpath)
    mydict["a"] = "b"
    new = persist(jsonpath)
    assert new.a == "b"


def test_persist_file_format(jsonpath):
    mydict = persist(jsonpath)
    mydict["a"] = "b"

    expected = '{\n  "a": "b"\n}'
    assert open(jsonpath, "r").read() == expected


def test_persist_get_attr_dict(jsonpath):
    mydict = persist(jsonpath)

    assert mydict._path == jsonpath
    assert mydict.__getattr__("_path") == jsonpath


def test_persist_set_attr_dict(jsonpath):
    mydict = persist(jsonpath)

    mydict["_path"] = "foo"
    mydict["a"] = "b"

    newdict = persist(jsonpath)
    assert newdict.a == "b"
